@@ -84,9 +84,9 @@ void MDSMonitor::create_initial()
 }
 
 
-void MDSMonitor::update_from_paxos()
+void MDSMonitor::update_from_paxos(bool *need_bootstrap)
 {
-  version_t version = get_version();
+  version_t version = get_last_committed();
   if (version == mdsmap.epoch)
     return;
   assert(version >= mdsmap.epoch);
@@ -127,7 +127,7 @@ void MDSMonitor::encode_pending(MonitorDBStore::Transaction *t)
   //print_map(pending_mdsmap);
 
   // apply to paxos
-  assert(get_version() + 1 == pending_mdsmap.epoch);
+  assert(get_last_committed() + 1 == pending_mdsmap.epoch);
   bufferlist mdsmap_bl;
   pending_mdsmap.encode(mdsmap_bl, mon->get_quorum_features());
 
@@ -529,6 +529,8 @@ void MDSMonitor::dump_info(Formatter *f)
   f->open_object_section("mdsmap");
   mdsmap.dump(f);
   f->close_section();
+
+  f->dump_unsigned("mdsmap_first_committed", get_first_committed());
 }
 
 bool MDSMonitor::preprocess_command(MMonCommand *m)
@@ -541,27 +543,35 @@ bool MDSMonitor::preprocess_command(MMonCommand *m)
   if (!cmdmap_from_json(m->cmd, &cmdmap, ss)) {
     // ss has reason for failure
     string rs = ss.str();
-    mon->reply_command(m, -EINVAL, rs, rdata, get_version());
+    mon->reply_command(m, -EINVAL, rs, rdata, get_last_committed());
     return true;
   }
 
   string prefix;
   cmd_getval(g_ceph_context, cmdmap, "prefix", prefix);
+  string format;
+  cmd_getval(g_ceph_context, cmdmap, "format", format, string("plain"));
+  boost::scoped_ptr<Formatter> f(new_formatter(format));
 
   MonSession *session = m->get_session();
   if (!session ||
       (!session->is_capable("mds", MON_CAP_R) &&
        !mon->_allowed_command(session, cmdmap))) {
-    mon->reply_command(m, -EACCES, "access denied", rdata, get_version());
+    mon->reply_command(m, -EACCES, "access denied", rdata, get_last_committed());
     return true;
   }
 
   if (prefix == "mds stat") {
-    ds << mdsmap;
+    if (f) {
+      f->open_object_section("mds_stat");
+      dump_info(f.get());
+      f->close_section();
+      f->flush(ds);
+    } else {
+      ds << mdsmap;
+    }
     r = 0;
   } else if (prefix == "mds dump") {
-    string format;
-    cmd_getval(g_ceph_context, cmdmap, "format", format, string("plain"));
     string val;
     int64_t epocharg;
     epoch_t epoch;
@@ -584,7 +594,6 @@ bool MDSMonitor::preprocess_command(MMonCommand *m)
     }
     if (p) {
       stringstream ds;
-      boost::scoped_ptr<Formatter> f(new_formatter(format));
       if (f != NULL) {
 	f->open_object_section("mdsmap");
 	p->dump(f.get());
@@ -638,7 +647,7 @@ bool MDSMonitor::preprocess_command(MMonCommand *m)
 	   i != mds_info.end();
 	   ++i) {
 	m->cmd = args_vec;
-	mon->send_command(i->second.get_inst(), m->cmd, get_version());
+	mon->send_command(i->second.get_inst(), m->cmd);
 	r = 0;
       }
       if (r == -ENOENT) {
@@ -652,7 +661,7 @@ bool MDSMonitor::preprocess_command(MMonCommand *m)
       if (!errno && who >= 0) {
 	if (mdsmap.is_up(who)) {
 	  m->cmd = args_vec;
-	  mon->send_command(mdsmap.get_inst(who), m->cmd, get_version());
+	  mon->send_command(mdsmap.get_inst(who), m->cmd);
 	  r = 0;
 	  ss << "ok";
 	} else {
@@ -662,7 +671,14 @@ bool MDSMonitor::preprocess_command(MMonCommand *m)
       } else ss << "specify mds number or *";
     }
   } else if (prefix == "mds compat show") {
-      ds << mdsmap.compat;
+      if (f) {
+	f->open_object_section("mds_compat");
+	mdsmap.compat.dump(f.get());
+	f->close_section();
+	f->flush(ds);
+      } else {
+	ds << mdsmap.compat;
+      }
       r = 0;
   }
 
@@ -670,7 +686,7 @@ bool MDSMonitor::preprocess_command(MMonCommand *m)
     rdata.append(ds);
     string rs;
     getline(ss, rs);
-    mon->reply_command(m, r, rs, rdata, get_version());
+    mon->reply_command(m, r, rs, rdata, get_last_committed());
     return true;
   } else
     return false;
@@ -734,43 +750,6 @@ int MDSMonitor::fail_mds(std::ostream &ss, const std::string &arg)
   return 0;
 }
 
-int MDSMonitor::cluster_fail(std::ostream &ss)
-{
-  dout(10) << "cluster_fail" << dendl;
-
-  if (!pending_mdsmap.test_flag(CEPH_MDSMAP_DOWN)) {
-    ss << "mdsmap must be marked DOWN first ('mds cluster_down')";
-    return -EPERM;
-  }
-  if (pending_mdsmap.up.size() && !mon->osdmon()->is_writeable()) {
-    ss << "osdmap not writeable, can't blacklist up mds's";
-    return -EAGAIN;
-  }
-
-  // --- reset the cluster map ---
-  if (pending_mdsmap.mds_info.size()) {
-    // blacklist all old mds's
-    utime_t until = ceph_clock_now(g_ceph_context);
-    until += g_conf->mds_blacklist_interval;
-    for (map<int32_t,uint64_t>::iterator p = pending_mdsmap.up.begin();
-	 p != pending_mdsmap.up.end();
-	 ++p) {
-      MDSMap::mds_info_t& info = pending_mdsmap.mds_info[p->second];
-      dout(10) << " blacklisting gid " << p->second << " " << info.addr << dendl;
-      pending_mdsmap.last_failure_osd_epoch = mon->osdmon()->blacklist(info.addr, until);
-    }
-    request_proposal(mon->osdmon());
-  }
-  pending_mdsmap.up.clear();
-  pending_mdsmap.failed.insert(pending_mdsmap.in.begin(),
-			       pending_mdsmap.in.end());
-  pending_mdsmap.in.clear();
-  pending_mdsmap.mds_info.clear();
-
-  ss << "failed all mds cluster members";
-  return 0;
-}
-
 bool MDSMonitor::prepare_command(MMonCommand *m)
 {
   int r = -EINVAL;
@@ -780,7 +759,7 @@ bool MDSMonitor::prepare_command(MMonCommand *m)
   map<string, cmd_vartype> cmdmap;
   if (!cmdmap_from_json(m->cmd, &cmdmap, ss)) {
     string rs = ss.str();
-    mon->reply_command(m, -EINVAL, rs, rdata, get_version());
+    mon->reply_command(m, -EINVAL, rs, rdata, get_last_committed());
     return true;
   }
 
@@ -792,7 +771,7 @@ bool MDSMonitor::prepare_command(MMonCommand *m)
   if (!session ||
       (!session->is_capable("mds", MON_CAP_W) &&
        !mon->_allowed_command(session, cmdmap))) {
-    mon->reply_command(m, -EACCES, "access denied", rdata, get_version());
+    mon->reply_command(m, -EACCES, "access denied", rdata, get_last_committed());
     return true;
   }
 
@@ -844,7 +823,7 @@ bool MDSMonitor::prepare_command(MMonCommand *m)
       map.epoch = pending_mdsmap.epoch;  // make sure epoch is correct
       pending_mdsmap = map;
       string rs = "set mds map";
-      wait_for_finished_proposal(new Monitor::C_Command(mon, m, 0, rs, get_version()));
+      wait_for_finished_proposal(new Monitor::C_Command(mon, m, 0, rs, get_last_committed()));
       return true;
     } else {
       ss << "next mdsmap epoch " << pending_mdsmap.epoch << " != " << e;
@@ -863,7 +842,7 @@ bool MDSMonitor::prepare_command(MMonCommand *m)
       ss << "set mds gid " << gid << " to state " << state << " " << ceph_mds_state_name(state);
       string rs;
       getline(ss, rs);
-      wait_for_finished_proposal(new Monitor::C_Command(mon, m, 0, rs, get_version()));
+      wait_for_finished_proposal(new Monitor::C_Command(mon, m, 0, rs, get_last_committed()));
       return true;
     }
 
@@ -893,7 +872,7 @@ bool MDSMonitor::prepare_command(MMonCommand *m)
       ss << "removed mds gid " << gid;
       string rs;
       getline(ss, rs);
-      wait_for_finished_proposal(new Monitor::C_Command(mon, m, 0, rs, get_version()));
+      wait_for_finished_proposal(new Monitor::C_Command(mon, m, 0, rs, get_last_committed()));
       return true;
     }
   } else if (prefix == "mds rmfailed") {
@@ -904,35 +883,24 @@ bool MDSMonitor::prepare_command(MMonCommand *m)
     ss << "removed failed mds." << w;
     string rs;
     getline(ss, rs);
-    wait_for_finished_proposal(new Monitor::C_Command(mon, m, 0, rs, get_version()));
+    wait_for_finished_proposal(new Monitor::C_Command(mon, m, 0, rs, get_last_committed()));
     return true;
-  } else if (prefix == "mds cluster_fail") {
-    r = cluster_fail(ss);
-    if (r < 0 && r == -EAGAIN) {
-      mon->osdmon()->wait_for_writeable(new C_RetryMessage(this,m));
-      return false; // don't propose yet; wait for message to be retried
-    }
-
   } else if (prefix == "mds cluster_down") {
     if (pending_mdsmap.test_flag(CEPH_MDSMAP_DOWN)) {
       ss << "mdsmap already marked DOWN";
-      r = -EPERM;
     } else {
       pending_mdsmap.set_flag(CEPH_MDSMAP_DOWN);
       ss << "marked mdsmap DOWN";
-      r = 0;
     }
-
+    r = 0;
   } else if (prefix == "mds cluster_up") {
     if (pending_mdsmap.test_flag(CEPH_MDSMAP_DOWN)) {
       pending_mdsmap.clear_flag(CEPH_MDSMAP_DOWN);
       ss << "unmarked mdsmap DOWN";
-      r = 0;
     } else {
       ss << "mdsmap not marked DOWN";
-      r = -EPERM;
     }
-
+    r = 0;
   } else if (prefix == "mds compat rm_compat") {
     int64_t f;
     cmd_getval(g_ceph_context, cmdmap, "feature", f);
@@ -967,6 +935,8 @@ bool MDSMonitor::prepare_command(MMonCommand *m)
     int64_t poolid;
     cmd_getval(g_ceph_context, cmdmap, "poolid", poolid);
     r = pending_mdsmap.remove_data_pool(poolid);
+    if (r == -ENOENT)
+      r = 0;
     if (r == 0)
       ss << "removed data pool " << poolid << " from mdsmap";
 
@@ -987,23 +957,23 @@ bool MDSMonitor::prepare_command(MMonCommand *m)
       ss << "new fs with metadata pool " << metadata << " and data pool " << data;
       string rs;
       getline(ss, rs);
-      wait_for_finished_proposal(new Monitor::C_Command(mon, m, 0, rs, get_version()));
+      wait_for_finished_proposal(new Monitor::C_Command(mon, m, 0, rs, get_last_committed()));
       return true;
     }
-  }    
-  if (r == -EINVAL) 
+  } else {
     ss << "unrecognized command";
+  }
  out:
   string rs;
   getline(ss, rs);
 
   if (r >= 0) {
     // success.. delay reply
-    wait_for_finished_proposal(new Monitor::C_Command(mon, m, r, rs, get_version()));
+    wait_for_finished_proposal(new Monitor::C_Command(mon, m, r, rs, get_last_committed()));
     return true;
   } else {
     // reply immediately
-    mon->reply_command(m, r, rs, rdata, get_version());
+    mon->reply_command(m, r, rs, rdata, get_last_committed());
     return false;
   }
 }
@@ -1040,7 +1010,6 @@ void MDSMonitor::tick()
   // ...if i am an active leader
   if (!is_active()) return;
 
-  update_from_paxos();
   dout(10) << mdsmap << dendl;
   
   bool do_propose = false;
