@@ -5,9 +5,10 @@
 
 #define dout_subsys ceph_subsys_mon
 #include "common/debug.h"
-
+#include "include/stringify.h"
 #include "common/Formatter.h"
 #include "include/ceph_features.h"
+#include "mon/MonitorDBStore.h"
 
 // --
 
@@ -109,7 +110,7 @@ void PGMap::Incremental::dump(Formatter *f) const
   f->close_section();
 
   f->open_array_section("osd_stat_updates");
-  for (map<int,osd_stat_t>::const_iterator p = osd_stat_updates.begin(); p != osd_stat_updates.end(); ++p) {
+  for (map<int32_t,osd_stat_t>::const_iterator p = osd_stat_updates.begin(); p != osd_stat_updates.end(); ++p) {
     f->open_object_section("osd_stat");
     f->dump_int("osd", p->first);
     p->second.dump(f);
@@ -176,6 +177,7 @@ void PGMap::apply_incremental(CephContext *cct, const Incremental& inc)
   }
   if (ratios_changed)
     redo_full_sets();
+
   for (map<pg_t,pg_stat_t>::const_iterator p = inc.pg_stat_updates.begin();
        p != inc.pg_stat_updates.end();
        ++p) {
@@ -192,15 +194,15 @@ void PGMap::apply_incremental(CephContext *cct, const Incremental& inc)
     }
     stat_pg_add(update_pg, update_stat);
   }
-  for (map<int,osd_stat_t>::const_iterator p = inc.osd_stat_updates.begin();
+  for (map<int32_t,osd_stat_t>::const_iterator p = inc.osd_stat_updates.begin();
        p != inc.osd_stat_updates.end();
        ++p) {
     int osd = p->first;
     const osd_stat_t &new_stats(p->second);
     
-    hash_map<int,osd_stat_t>::iterator t = osd_stat.find(osd);
+    hash_map<int32_t,osd_stat_t>::iterator t = osd_stat.find(osd);
     if (t == osd_stat.end()) {
-      hash_map<int,osd_stat_t>::value_type v(osd, new_stats);
+      hash_map<int32_t,osd_stat_t>::value_type v(osd, new_stats);
       osd_stat.insert(v);
     } else {
       stat_osd_sub(t->second);
@@ -226,7 +228,7 @@ void PGMap::apply_incremental(CephContext *cct, const Incremental& inc)
   for (set<int>::iterator p = inc.osd_stat_rm.begin();
        p != inc.osd_stat_rm.end();
        ++p) {
-    hash_map<int,osd_stat_t>::iterator t = osd_stat.find(*p);
+    hash_map<int32_t,osd_stat_t>::iterator t = osd_stat.find(*p);
     if (t != osd_stat.end()) {
       stat_osd_sub(t->second);
       osd_stat.erase(t);
@@ -260,7 +262,7 @@ void PGMap::redo_full_sets()
 {
   full_osds.clear();
   nearfull_osds.clear();
-  for (hash_map<int, osd_stat_t>::iterator i = osd_stat.begin();
+  for (hash_map<int32_t, osd_stat_t>::iterator i = osd_stat.begin();
        i != osd_stat.end();
        ++i) {
     register_nearfull_status(i->first, i->second);
@@ -300,7 +302,7 @@ void PGMap::calc_stats()
        ++p) {
     stat_pg_add(p->first, p->second);
   }
-  for (hash_map<int,osd_stat_t>::iterator p = osd_stat.begin();
+  for (hash_map<int32_t,osd_stat_t>::iterator p = osd_stat.begin();
        p != osd_stat.end();
        ++p)
     stat_osd_add(p->second);
@@ -308,6 +310,53 @@ void PGMap::calc_stats()
   redo_full_sets();
 }
 
+void PGMap::update_pg(pg_t pgid, bufferlist& bl)
+{
+  bufferlist::iterator p = bl.begin();
+  hash_map<pg_t,pg_stat_t>::iterator s = pg_stat.find(pgid);
+  if (s != pg_stat.end())
+    stat_pg_sub(pgid, s->second);
+  pg_stat_t& r = pg_stat[pgid];
+  ::decode(r, p);
+  stat_pg_add(pgid, r);
+}
+
+void PGMap::remove_pg(pg_t pgid)
+{
+  hash_map<pg_t,pg_stat_t>::iterator s = pg_stat.find(pgid);
+  if (s != pg_stat.end()) {
+    stat_pg_sub(pgid, s->second);
+    pg_stat.erase(s);
+  }
+}
+
+void PGMap::update_osd(int osd, bufferlist& bl)
+{
+  bufferlist::iterator p = bl.begin();
+  hash_map<int32_t,osd_stat_t>::iterator o = osd_stat.find(osd);
+  if (o != osd_stat.end())
+    stat_osd_sub(o->second);
+  osd_stat_t& r = osd_stat[osd];
+  ::decode(r, p);
+  stat_osd_add(r);
+
+  // adjust [near]full status
+  register_nearfull_status(osd, r);
+}
+
+void PGMap::remove_osd(int osd)
+{
+  hash_map<int32_t,osd_stat_t>::iterator o = osd_stat.find(osd);
+  if (o != osd_stat.end()) {
+    stat_osd_sub(o->second);
+    osd_stat.erase(o);
+
+    // remove these old osds from full/nearfull set(s), too
+    nearfull_osds.erase(osd);
+    full_osds.erase(osd);
+  }
+}
+
 void PGMap::stat_pg_add(const pg_t &pgid, const pg_stat_t &s)
 {
   num_pg++;
@@ -360,10 +409,11 @@ epoch_t PGMap::calc_min_last_epoch_clean() const
   if (pg_stat.empty())
     return 0;
   hash_map<pg_t,pg_stat_t>::const_iterator p = pg_stat.begin();
-  epoch_t min = p->second.last_epoch_clean;
+  epoch_t min = p->second.get_effective_last_epoch_clean();
   for (++p; p != pg_stat.end(); ++p) {
-    if (p->second.last_epoch_clean < min)
-      min = p->second.last_epoch_clean;
+    epoch_t lec = p->second.get_effective_last_epoch_clean();
+    if (lec < min)
+      min = lec;
   }
   return min;
 }
@@ -426,10 +476,25 @@ void PGMap::decode(bufferlist::iterator &bl)
   calc_stats();
 }
 
+void PGMap::dirty_all(Incremental& inc)
+{
+  inc.osdmap_epoch = last_osdmap_epoch;
+  inc.pg_scan = last_pg_scan;
+  inc.full_ratio = full_ratio;
+  inc.nearfull_ratio = nearfull_ratio;
+
+  for (hash_map<pg_t,pg_stat_t>::const_iterator p = pg_stat.begin(); p != pg_stat.end(); ++p) {
+    inc.pg_stat_updates[p->first] = p->second;
+  }
+  for (hash_map<int32_t, osd_stat_t>::const_iterator p = osd_stat.begin(); p != osd_stat.end(); ++p) {
+    inc.osd_stat_updates[p->first] = p->second;
+  }
+}
+
 void PGMap::dump(Formatter *f) const
 {
   dump_basic(f);
-  dump_pg_stats(f);
+  dump_pg_stats(f, false);
   dump_pool_stats(f);
   dump_osd_stats(f);
 }
@@ -456,7 +521,7 @@ void PGMap::dump_basic(Formatter *f) const
   f->close_section();
 }
 
-void PGMap::dump_pg_stats(Formatter *f) const
+void PGMap::dump_pg_stats(Formatter *f, bool brief) const
 {
   f->open_array_section("pg_stats");
   for (hash_map<pg_t,pg_stat_t>::const_iterator i = pg_stat.begin();
@@ -464,7 +529,10 @@ void PGMap::dump_pg_stats(Formatter *f) const
        ++i) {
     f->open_object_section("pg_stat");
     f->dump_stream("pgid") << i->first;
-    i->second.dump(f);
+    if (brief)
+      i->second.dump_brief(f);
+    else
+      i->second.dump(f);
     f->close_section();
   }
   f->close_section();
@@ -487,7 +555,7 @@ void PGMap::dump_pool_stats(Formatter *f) const
 void PGMap::dump_osd_stats(Formatter *f) const
 {
   f->open_array_section("osd_stats");
-  for (hash_map<int,osd_stat_t>::const_iterator q = osd_stat.begin();
+  for (hash_map<int32_t,osd_stat_t>::const_iterator q = osd_stat.begin();
        q != osd_stat.end();
        ++q) {
     f->open_object_section("osd_stat");
@@ -517,7 +585,7 @@ void PGMap::dump_pg_stats_plain(ostream& ss,
        << "\t" << pg_state_string(st.state)
        << "\t" << st.last_change
        << "\t" << st.version
-       << "\t" << st.reported
+       << "\t" << st.reported_epoch << ":" << st.reported_seq
        << "\t" << st.up
        << "\t" << st.acting
        << "\t" << st.last_scrub << "\t" << st.last_scrub_stamp
@@ -558,7 +626,7 @@ void PGMap::dump(ostream& ss) const
      << "\t" << pg_sum.ondisk_log_size
      << std::endl;
   ss << "osdstat\tkbused\tkbavail\tkb\thb in\thb out" << std::endl;
-  for (hash_map<int,osd_stat_t>::const_iterator p = osd_stat.begin();
+  for (hash_map<int32_t,osd_stat_t>::const_iterator p = osd_stat.begin();
        p != osd_stat.end();
        ++p)
     ss << p->first
@@ -630,48 +698,85 @@ void PGMap::dump_stuck_plain(ostream& ss, PGMap::StuckPG type, utime_t cutoff) c
   dump_pg_stats_plain(ss, stuck_pg_stats);
 }
 
-void PGMap::state_summary(ostream& ss) const
-{
-  for (hash_map<int,int>::const_iterator p = num_pg_by_state.begin();
-       p != num_pg_by_state.end();
-       ++p) {
-    if (p != num_pg_by_state.begin())
-      ss << ", ";
-    ss << p->second << " " << pg_state_string(p->first);
-  }
-}
-
-void PGMap::recovery_summary(ostream& out) const
+void PGMap::recovery_summary(Formatter *f, ostream *out) const
 {
   bool first = true;
   if (pg_sum.stats.sum.num_objects_degraded) {
     double pc = (double)pg_sum.stats.sum.num_objects_degraded / (double)pg_sum.stats.sum.num_object_copies * (double)100.0;
     char b[20];
     snprintf(b, sizeof(b), "%.3lf", pc);
-    out << pg_sum.stats.sum.num_objects_degraded 
-	<< "/" << pg_sum.stats.sum.num_object_copies << " degraded (" << b << "%)";
+    if (f) {
+      f->dump_unsigned("degraded_objects", pg_sum.stats.sum.num_objects_degraded);
+      f->dump_unsigned("degraded_total", pg_sum.stats.sum.num_object_copies);
+      f->dump_string("degrated_ratio", b);
+    } else {
+      *out << pg_sum.stats.sum.num_objects_degraded 
+	   << "/" << pg_sum.stats.sum.num_object_copies << " degraded (" << b << "%)";
+    }
     first = false;
   }
   if (pg_sum.stats.sum.num_objects_unfound) {
     double pc = (double)pg_sum.stats.sum.num_objects_unfound / (double)pg_sum.stats.sum.num_objects * (double)100.0;
     char b[20];
     snprintf(b, sizeof(b), "%.3lf", pc);
-    if (!first)
-      out << "; ";
-    out << pg_sum.stats.sum.num_objects_unfound
-	<< "/" << pg_sum.stats.sum.num_objects << " unfound (" << b << "%)";
+    if (f) {
+      f->dump_unsigned("unfound_objects", pg_sum.stats.sum.num_objects_unfound);
+      f->dump_unsigned("unfound_total", pg_sum.stats.sum.num_objects);
+      f->dump_string("unfound_ratio", b);
+    } else {
+      if (!first)
+	*out << "; ";
+      *out << pg_sum.stats.sum.num_objects_unfound
+	   << "/" << pg_sum.stats.sum.num_objects << " unfound (" << b << "%)";
+    }
     first = false;
   }
-  if (pg_sum_delta.stats.sum.num_objects_recovered ||
-      pg_sum_delta.stats.sum.num_bytes_recovered ||
-      pg_sum_delta.stats.sum.num_keys_recovered) {
-    if (!first)
-      out << "; ";
-    out << " recovering "
-	<< si_t(pg_sum_delta.stats.sum.num_objects_recovered / (double)stamp_delta) << " o/s, "
-	<< si_t(pg_sum_delta.stats.sum.num_bytes_recovered / (double)stamp_delta) << "B/s";
-    if (pg_sum_delta.stats.sum.num_keys_recovered)
-      out << ", " << si_t(pg_sum_delta.stats.sum.num_keys_recovered / (double)stamp_delta) << " key/s";
+
+  // make non-negative; we can get negative values if osds send
+  // uncommitted stats and then "go backward" or if they are just
+  // buggy/wrong.
+  pool_stat_t pos_delta = pg_sum_delta;
+  pos_delta.floor(0);
+  if (pos_delta.stats.sum.num_objects_recovered ||
+      pos_delta.stats.sum.num_bytes_recovered ||
+      pos_delta.stats.sum.num_keys_recovered) {
+    int64_t objps = pos_delta.stats.sum.num_objects_recovered / (double)stamp_delta;
+    int64_t bps = pos_delta.stats.sum.num_bytes_recovered / (double)stamp_delta;
+    int64_t kps = pos_delta.stats.sum.num_keys_recovered / (double)stamp_delta;
+    if (f) {
+      f->dump_int("recovering_objects_per_sec", objps);
+      f->dump_int("recovering_bytes_per_sec", bps);
+      f->dump_int("recovering_keys_per_sec", kps);
+    } else {
+      if (!first)
+	*out << "; ";
+      *out << " recovering "
+	   << si_t(objps) << " o/s, "
+	   << si_t(bps) << "B/s";
+      if (pos_delta.stats.sum.num_keys_recovered)
+	*out << ", " << si_t(kps) << " key/s";
+    }
+  }
+}
+
+void PGMap::update_delta(CephContext *cct, utime_t inc_stamp, pool_stat_t& pg_sum_old)
+{
+  utime_t delta_t;
+  delta_t = inc_stamp;
+  delta_t -= stamp;
+  stamp = inc_stamp;
+
+  // calculate a delta, and average over the last 2 deltas.
+  pool_stat_t d = pg_sum;
+  d.stats.sub(pg_sum_old.stats);
+  pg_sum_deltas.push_back(make_pair(d, delta_t));
+  stamp_delta += delta_t;
+
+  pg_sum_delta.stats.add(d.stats);
+  if (pg_sum_deltas.size() > (std::list< pair<pool_stat_t, utime_t> >::size_type)MAX(1, cct ? cct->_conf->mon_stat_smooth_intervals : 1)) {
+    pg_sum_delta.stats.sub(pg_sum_deltas.front().first.stats);
+    stamp_delta -= pg_sum_deltas.front().second;
+    pg_sum_deltas.pop_front();
   }
 }
 
@@ -679,36 +784,87 @@ void PGMap::clear_delta()
 {
   pg_sum_delta = pool_stat_t();
   pg_sum_deltas.clear();
-  stamp_delta = ceph_clock_now(g_ceph_context);
+  stamp_delta = utime_t();
 }
 
-void PGMap::print_summary(ostream& out) const
+void PGMap::print_summary(Formatter *f, ostream *out) const
 {
   std::stringstream ss;
-  state_summary(ss);
-  string states = ss.str();
-  out << "v" << version << ": "
-      << pg_stat.size() << " pgs: "
-      << states << "; "
-      << prettybyte_t(pg_sum.stats.sum.num_bytes) << " data, " 
-      << kb_t(osd_sum.kb_used) << " used, "
-      << kb_t(osd_sum.kb_avail) << " / "
-      << kb_t(osd_sum.kb) << " avail";
+  if (f)
+    f->open_array_section("pgs_by_state");
 
-  if (pg_sum_delta.stats.sum.num_rd ||
-      pg_sum_delta.stats.sum.num_wr) {
-    out << "; ";
-    if (pg_sum_delta.stats.sum.num_rd)
-      out << si_t((pg_sum_delta.stats.sum.num_rd_kb << 10) / (double)stamp_delta) << "B/s rd, ";
-    if (pg_sum_delta.stats.sum.num_wr)
-      out << si_t((pg_sum_delta.stats.sum.num_wr_kb << 10) / (double)stamp_delta) << "B/s wr, ";
-    out << si_t((pg_sum_delta.stats.sum.num_rd + pg_sum_delta.stats.sum.num_wr) / (double)stamp_delta) << "op/s";
+  for (hash_map<int,int>::const_iterator p = num_pg_by_state.begin();
+       p != num_pg_by_state.end();
+       ++p) {
+    if (f) {
+      f->open_object_section("pgs_by_state_element");
+      f->dump_string("state_name", pg_state_string(p->first));
+      f->dump_unsigned("count", p->second);
+      f->close_section();
+    } else {
+      if (p != num_pg_by_state.begin())
+	ss << ", ";
+      ss << p->second << " " << pg_state_string(p->first);
+    }
+  }
+  if (f)
+    f->close_section();
+
+  if (f) {
+    f->dump_unsigned("version", version);
+    f->dump_unsigned("num_pgs", pg_stat.size());
+    f->dump_unsigned("data_bytes", pg_sum.stats.sum.num_bytes);
+    f->dump_unsigned("bytes_used", osd_sum.kb_used * 4096ull);
+    f->dump_unsigned("bytes_avail", osd_sum.kb_avail * 4096ull);
+    f->dump_unsigned("bytes_total", osd_sum.kb * 4096ull);
+  } else {
+    string states = ss.str();
+    *out << "v" << version << ": "
+	 << pg_stat.size() << " pgs: "
+	 << states << "; "
+	 << prettybyte_t(pg_sum.stats.sum.num_bytes) << " data, " 
+	 << kb_t(osd_sum.kb_used) << " used, "
+	 << kb_t(osd_sum.kb_avail) << " / "
+	 << kb_t(osd_sum.kb) << " avail";
+  }
+
+  // make non-negative; we can get negative values if osds send
+  // uncommitted stats and then "go backward" or if they are just
+  // buggy/wrong.
+  pool_stat_t pos_delta = pg_sum_delta;
+  pos_delta.floor(0);
+  if (pos_delta.stats.sum.num_rd ||
+      pos_delta.stats.sum.num_wr) {
+    if (!f)
+      *out << "; ";
+    if (pos_delta.stats.sum.num_rd) {
+      int64_t rd = (pos_delta.stats.sum.num_rd_kb << 10) / (double)stamp_delta;
+      if (f) {
+	f->dump_int("read_bytes_sec", rd);
+      } else {
+	*out << si_t(rd) << "B/s rd, ";
+      }
+    }
+    if (pos_delta.stats.sum.num_wr) {
+      int64_t wr = (pos_delta.stats.sum.num_wr_kb << 10) / (double)stamp_delta;
+      if (f) {
+	f->dump_int("write_bytes_sec", wr);
+      } else {
+	*out << si_t(wr) << "B/s wr, ";
+      }
+    }
+    int64_t iops = (pos_delta.stats.sum.num_rd + pos_delta.stats.sum.num_wr) / (double)stamp_delta;
+    if (f) {
+      f->dump_int("op_per_sec", iops);
+    } else {
+      *out << si_t(iops) << "op/s";
+    }
   }
 
   std::stringstream ssr;
-  recovery_summary(ssr);
-  if (ssr.str().length())
-    out << "; " << ssr.str();
+  recovery_summary(f, &ssr);
+  if (!f && ssr.str().length())
+    *out << "; " << ssr.str();
 }
 
 void PGMap::generate_test_instances(list<PGMap*>& o)
