@@ -31,10 +31,81 @@ b 12v
 c 14v
 d
 e 12v
-
-
 */
 
+/**
+ * Paxos storage layout and behavior
+ *
+ * Currently, we use a key/value store to hold all the Paxos-related data, but
+ * it can logically be depicted as this:
+ *
+ *  paxos:
+ *    first_committed -> 1
+ *     last_committed -> 4
+ *		    1 -> value_1
+ *		    2 -> value_2
+ *		    3 -> value_3
+ *		    4 -> value_4
+ *
+ * Since we are relying on a k/v store supporting atomic transactions, we can
+ * guarantee that if 'last_committed' has a value of '4', then we have up to
+ * version 4 on the store, and no more than that; the same applies to
+ * 'first_committed', which holding '1' will strictly meaning that our lowest
+ * version is 1.
+ *
+ * Each version's value (value_1, value_2, ..., value_n) is a blob of data,
+ * incomprehensible to the Paxos. These values are proposed to the Paxos on
+ * propose_new_value() and each one is a transaction encoded in a bufferlist.
+ *
+ * The Paxos will write the value to disk, associating it with its version,
+ * but will take a step further: the value shall be decoded, and the operations
+ * on that transaction shall be applied during the same transaction that will
+ * write the value's encoded bufferlist to disk. This behavior ensures that
+ * whatever is being proposed will only be available on the store when it is
+ * applied by Paxos, which will then be aware of such new values, guaranteeing
+ * the store state is always consistent without requiring shady workarounds.
+ *
+ * So, let's say that FooMonitor proposes the following transaction, neatly
+ * encoded on a bufferlist of course:
+ *
+ *  Tx_Foo
+ *    put(foo, last_committed, 3)
+ *    put(foo, 3, foo_value_3)
+ *    erase(foo, 2)
+ *    erase(foo, 1)
+ *    put(foo, first_committed, 3)
+ *
+ * And knowing that the Paxos is proposed Tx_Foo as a bufferlist, once it is
+ * ready to commit, and assuming we are now committing version 5 of the Paxos,
+ * we will do something along the lines of:
+ *
+ *  Tx proposed_tx;
+ *  proposed_tx.decode(Tx_foo_bufferlist);
+ *
+ *  Tx our_tx;
+ *  our_tx.put(paxos, last_committed, 5);
+ *  our_tx.put(paxos, 5, Tx_foo_bufferlist);
+ *  our_tx.append(proposed_tx);
+ *
+ *  store_apply(our_tx);
+ *
+ * And the store should look like this after we apply 'our_tx':
+ *
+ *  paxos:
+ *    first_committed -> 1
+ *     last_committed -> 5
+ *		    1 -> value_1
+ *		    2 -> value_2
+ *		    3 -> value_3
+ *		    4 -> value_4
+ *		    5 -> Tx_foo_bufferlist
+ *  foo:
+ *    first_committed -> 3
+ *     last_committed -> 3
+ *		    3 -> foo_value_3
+ *
+ */
+
 #ifndef CEPH_MON_PAXOS_H
 #define CEPH_MON_PAXOS_H
 
@@ -49,11 +120,12 @@ e 12v
 #include "common/Timer.h"
 #include <errno.h>
 
+#include "MonitorDBStore.h"
+
 class Monitor;
 class MMonPaxos;
 class Paxos;
 
-
 // i am one state machine.
 /**
  * This libary is based on the Paxos algorithm, but varies in a few key ways:
@@ -76,8 +148,7 @@ class Paxos {
   Monitor *mon;
 
   // my state machine info
-  int machine_id;
-  const char *machine_name;
+  const string paxos_name;
 
   friend class Monitor;
   friend class PaxosService;
@@ -92,19 +163,25 @@ public:
    * @defgroup Paxos_h_states States on which the leader/peon may be.
    * @{
    */
-  /**
-   * Leader/Peon is in Paxos' Recovery state
-   */
-  const static int STATE_RECOVERING = 1;
-  /**
-   * Leader/Peon is idle, and the Peon may or may not have a valid lease.
-   */
-  const static int STATE_ACTIVE     = 2;
-  /**
-   * Leader/Peon is updating to a new value.
-   */
-  const static int STATE_UPDATING   = 3;
- 
+  enum {
+    /**
+     * Leader/Peon is in Paxos' Recovery state
+     */
+    STATE_RECOVERING,
+    /**
+     * Leader/Peon is idle, and the Peon may or may not have a valid lease.
+     */
+    STATE_ACTIVE,
+    /**
+     * Leader/Peon is updating to a new value.
+     */
+    STATE_UPDATING,
+    /*
+     * Leader proposing an old value
+     */
+    STATE_UPDATING_PREVIOUS,
+  };
+
   /**
    * Obtain state name from constant value.
    *
@@ -114,12 +191,18 @@ public:
    * @param s State value.
    * @return The state's name.
    */
-  static const char *get_statename(int s) {
+  static const string get_statename(int s) {
     switch (s) {
-    case STATE_RECOVERING: return "recovering";
-    case STATE_ACTIVE: return "active";
-    case STATE_UPDATING: return "updating";
-    default: assert(0); return 0;
+    case STATE_RECOVERING:
+      return "recovering";
+    case STATE_ACTIVE:
+      return "active";
+    case STATE_UPDATING:
+      return "updating";
+    case STATE_UPDATING_PREVIOUS:
+      return "updating-previous";
+    default:
+      return "UNKNOWN";
     }
   }
 
@@ -138,7 +221,7 @@ public:
    *
    * @return 'true' if we are on the Recovering state; 'false' otherwise.
    */
-  bool is_recovering() const { return state == STATE_RECOVERING; }
+  bool is_recovering() const { return (state & STATE_RECOVERING); }
   /**
    * Check if we are active.
    *
@@ -152,6 +235,12 @@ public:
    */
   bool is_updating() const { return state == STATE_UPDATING; }
 
+  /**
+   * Check if we are updating/proposing a previous value from a
+   * previous quorum
+   */
+  bool is_updating_previous() const { return state == STATE_UPDATING_PREVIOUS; }
+
 private:
   /**
    * @defgroup Paxos_h_recovery_vars Common recovery-related member variables
@@ -201,13 +290,17 @@ private:
    */
   version_t accepted_pn;
   /**
-   * @todo This has something to do with the last_committed version. Not sure
-   *	   about what it entails, tbh.
+   * The last_committed epoch of the leader at the time we accepted the last pn.
+   *
+   * This has NO SEMANTIC MEANING, and is there only for the debug output.
    */
   version_t accepted_pn_from;
   /**
-   * @todo Check out if this map has any purpose at all. So far, we have only
-   *	   seen it being read, although it is never affected.
+   * Map holding the first committed version by each quorum member.
+   *
+   * The versions kept in this map are updated during the collect phase.
+   * When the Leader starts the collect phase, each Peon will reply with its
+   * first committed version, which will then be kept in this map.
    */
   map<int,version_t> peer_first_committed;
   /**
@@ -218,10 +311,6 @@ private:
    * last committed version, which will then be kept in this map.
    */
   map<int,version_t> peer_last_committed;
-  /**
-   * @todo Check out what 'slurping' is.
-   */
-  int slurping;
   /**
    * @}
    */
@@ -260,11 +349,6 @@ private:
    * uncommitted values --, or if we're on a quorum of one.
    */
   list<Context*> waiting_for_readable;
-
-  /**
-   * Latest version written to the store after the latest commit.
-   */
-  version_t latest_stashed;
   /**
    * @}
    */
@@ -423,6 +507,10 @@ private:
    *	      fully committed.
    */
   list<Context*> waiting_for_commit;
+  /**
+   *
+   */
+  list<Context*> proposals;
   /**
    * @}
    */
@@ -438,6 +526,11 @@ private:
    * @}
    */
 
+  /**
+   * Should be true if we have proposed to trim, or are in the middle of
+   * trimming; false otherwise.
+   */
+  bool trimming;
 
   /**
    * @defgroup Paxos_h_callbacks Callback classes.
@@ -512,10 +605,45 @@ private:
       paxos->lease_renew_timeout();
     }
   };
+
+  class C_Trimmed : public Context {
+    Paxos *paxos;
+  public:
+    C_Trimmed(Paxos *p) : paxos(p) { }
+    void finish(int r) {
+      paxos->trimming = false;
+    }
+  };
+  /**
+   *
+   */
+public:
+  class C_Proposal : public Context {
+    Context *proposer_context;
+  public:
+    bufferlist bl;
+    // for debug purposes. Will go away. Soon.
+    bool proposed;
+    utime_t proposal_time;
+
+    C_Proposal(Context *c, bufferlist& proposal_bl) :
+	proposer_context(c),
+	bl(proposal_bl),
+        proposed(false),
+	proposal_time(ceph_clock_now(NULL))
+      { }
+
+    void finish(int r) {
+      if (proposer_context) {
+	proposer_context->complete(r);
+	proposer_context = NULL;
+      }
+    }
+  };
   /**
    * @}
    */
-
+private:
   /**
    * @defgroup Paxos_h_election_triggered Steps triggered by an election.
    *
@@ -634,7 +762,7 @@ private:
    *
    * @param value The value being proposed to the quorum
    */
-  void begin(bufferlist& value, version_t global_version);
+  void begin(bufferlist& value);
   /**
    * Accept or decline (by ignoring) a proposal from the Leader.
    *
@@ -818,6 +946,9 @@ private:
    */
   void lease_timeout();        // on peon, if lease isn't extended
 
+  /// restart the lease timeout timer
+  void reset_lease_timeout();
+
   /**
    * Cancel all of Paxos' timeout/renew events. 
    */
@@ -841,23 +972,46 @@ private:
    */
   void warn_on_future_time(utime_t t, entity_name_t from);
 
+  /**
+   * Queue a new proposal by pushing it at the back of the queue; do not
+   * propose it.
+   *
+   * @param bl The bufferlist to be proposed
+   * @param onfinished The callback to be called once the proposal finishes
+   */
+  void queue_proposal(bufferlist& bl, Context *onfinished);
+  /**
+   * Begin proposing the Proposal at the front of the proposals queue.
+   */
+  void propose_queued();
+
+  /**
+   * refresh state from store
+   *
+   * Called when we have new state for the mon to consume.  If we return false,
+   * abort (we triggered a bootstrap).
+   *
+   * @returns true on success, false if we are now bootstrapping
+   */
+  bool do_refresh();
+
+  void commit_proposal();
+  void finish_round();
+
 public:
   /**
    * @param m A monitor
    * @param mid A machine id
    */
-  Paxos(Monitor *m,
-	int mid) : mon(m),
-		   machine_id(mid), 
-		   machine_name(get_paxos_name(mid)),
+  Paxos(Monitor *m, const string &name) 
+		 : mon(m),
+		   paxos_name(name),
 		   state(STATE_RECOVERING),
 		   first_committed(0),
 		   last_pn(0),
 		   last_committed(0),
 		   accepted_pn(0),
 		   accepted_pn_from(0),
-		   slurping(0),
-		   latest_stashed(0),
 		   num_last(0),
 		   uncommitted_v(0), uncommitted_pn(0),
 		   collect_timeout_event(0),
@@ -865,15 +1019,24 @@ public:
 		   lease_ack_timeout_event(0),
 		   lease_timeout_event(0),
 		   accept_timeout_event(0),
-		   clock_drift_warned(0) { }
+		   clock_drift_warned(0),
+		   trimming(false) { }
 
-  const char *get_machine_name() const {
-    return machine_name;
+  const string get_name() const {
+    return paxos_name;
   }
 
   void dispatch(PaxosServiceMessage *m);
 
+  void read_and_prepare_transactions(MonitorDBStore::Transaction *tx, version_t from, version_t last);
+
   void init();
+
+  /**
+   * dump state info to a formatter
+   */
+  void dump_info(Formatter *f);
+
   /**
    * This function runs basic consistency checks. Importantly, if
    * it is inconsistent and shouldn't be, it asserts out.
@@ -920,11 +1083,45 @@ public:
   void share_state(MMonPaxos *m, version_t peer_first_committed,
 		   version_t peer_last_committed);
   /**
-   * Store the state held on the message m into local, stable storage.
+   * Store on disk a state that was shared with us
+   *
+   * Basically, we received a set of version. Or just one. It doesn't matter.
+   * What matters is that we have to stash it in the store. So, we will simply
+   * write every single bufferlist into their own versions on our side (i.e.,
+   * onto paxos-related keys), and then we will decode those same bufferlists
+   * we just wrote and apply the transactions they hold. We will also update
+   * our first and last committed values to point to the new values, if need
+   * be. All all this is done tightly wrapped in a transaction to ensure we
+   * enjoy the atomicity guarantees given by our awesome k/v store.
    *
    * @param m A message
    */
   void store_state(MMonPaxos *m);
+  void _sanity_check_store();
+
+  /**
+   * remove legacy paxos versions from before conversion
+   */
+  void remove_legacy_versions();
+
+  /**
+   * Helper function to decode a bufferlist into a transaction and append it
+   * to another transaction.
+   *
+   * This function is used during the Leader's commit and during the
+   * Paxos::store_state in order to apply the bufferlist's transaction onto
+   * the store.
+   *
+   * @param t The transaction to which we will append the operations
+   * @param bl A bufferlist containing an encoded transaction
+   */
+  static void decode_append_transaction(MonitorDBStore::Transaction& t,
+				 bufferlist& bl) {
+    MonitorDBStore::Transaction vt;
+    bufferlist::iterator it = bl.begin();
+    vt.decode(it);
+    t.append(vt);
+  }
 
   /**
    * @todo This appears to be used only by the OSDMonitor, and I would say
@@ -946,26 +1143,28 @@ public:
   }
 
   /**
-   * Erase old states from stable storage.
-   *
-   * @param first The version we are trimming to
-   * @param force If specified, we may even erase the latest stashed version
-   *		  iif @p first is higher than that version.
-   */
-  void trim_to(version_t first, bool force=false);
- 
-  /**
-   * @defgroup Paxos_h_slurping_funcs Slurping-related functions
-   * @todo Discover what slurping is
-   * @{
-   */
-  void start_slurping();
-  void end_slurping();
-  bool is_slurping() { return slurping == 1; }
-  /**
-   * @}
+   * Trim the Paxos state as much as we can.
    */
+  void trim();
 
+  /**
+   * Check if we should trim.
+   *
+   * If trimming is disabled, we must take that into consideration and only
+   * return true if we are positively sure that we should trim soon.
+   *
+   * @returns true if we should trim; false otherwise.
+   */
+  bool should_trim() {
+    int available_versions = get_version() - get_first_committed();
+    int maximum_versions = g_conf->paxos_min + g_conf->paxos_trim_min;
+
+    if (trimming || (available_versions <= maximum_versions))
+      return false;
+
+    return true;
+  }
+ 
   // read
   /**
    * @defgroup Paxos_h_read_funcs Read-related functions
@@ -1018,7 +1217,7 @@ public:
    * @param onreadable A callback
    */
   void wait_for_readable(Context *onreadable) {
-    //assert(!is_readable());
+    assert(!is_readable());
     waiting_for_readable.push_back(onreadable);
   }
   /**
@@ -1026,10 +1225,11 @@ public:
    */
 
   /**
-   * @warning This declaration is not implemented anywhere and appears to be
-   *	      just some lingering code.
+   * Check if we have a valid lease.
+   *
+   * @returns true if the lease is still valid; false otherwise.
    */
-  bool is_leader();
+  bool is_lease_valid();
   // write
   /**
    * @defgroup Paxos_h_write_funcs Write-related functions
@@ -1057,6 +1257,13 @@ public:
     waiting_for_writeable.push_back(c);
   }
 
+  /**
+   * List all queued proposals
+   *
+   * @param out[out] Output Stream onto which we will output the list
+   *		     of queued proposals.
+   */
+  void list_proposals(ostream& out);
   /**
    * Propose a new value to the Leader.
    *
@@ -1064,9 +1271,9 @@ public:
    * will trigger a new proposal.
    *
    * @param bl A bufferlist holding the value to be proposed
-   * @param oncommit A callback to be fired up once we finish committing bl
+   * @param onfinish A callback to be fired up once we finish the proposal
    */
-  bool propose_new_value(bufferlist& bl, Context *oncommit=0);
+  bool propose_new_value(bufferlist& bl, Context *onfinished=0);
   /**
    * Add oncommit to the back of the list of callbacks waiting for us to
    * finish committing.
@@ -1089,47 +1296,27 @@ public:
    * @}
    */
 
-  /**
-   * @defgroup Paxos_h_stash_funcs State values stashing-related functions
-   *
-   * If the state values are incrementals, it is useful to keep the latest
-   * copy of the complete structure.
-   *
-   * @{
-   */
-  /**
-   * Get the latest version onto stable storage.
-   *
-   * Keeping the latest version on a predefined location makes it easier to
-   * access, since we know we always have the latest version on the same
-   * place.
-   *
-   * @param v the latest version
-   * @param bl the latest version's value
-   */
-  void stash_latest(version_t v, bufferlist& bl);
-  /**
-   * Get the latest stashed version's value
-   *
-   * @param[out] bl the latest stashed version's value
-   * @return the latest stashed version
-   */
-  version_t get_stashed(bufferlist& bl);
-  /**
-   * Get the latest stashed version
-   *
-   * @return the latest stashed version
-   */
-  version_t get_stashed_version() { return latest_stashed; }
-  /**
-   * @}
-   */
-
   /**
    * @}
    */
+ protected:
+  MonitorDBStore *get_store();
 };
 
+inline ostream& operator<<(ostream& out, Paxos::C_Proposal& p)
+{
+  string proposed = (p.proposed ? "proposed" : "unproposed");
+  out << " " << proposed
+      << " queued " << (ceph_clock_now(NULL) - p.proposal_time)
+      << " tx dump:\n";
+  MonitorDBStore::Transaction t;
+  bufferlist::iterator p_it = p.bl.begin();
+  t.decode(p_it);
+  JSONFormatter f(true);
+  t.dump(&f);
+  f.flush(out);
+  return out;
+}
 
 #endif
 
